#include <stdlib.h>
#include "./include/gc.h"
// #include "./internal/obj.h"

static uint64_t bucket_i(size_t cap, void *ptr) {
	return (((size_t)ptr / 256) * 4567) % cap;
}

static void table_enlarge(js_gc_table_t *table) {
	js_gc_table_t new_table = {
		.cap = table->cap * 2,
		.n = table->n,
		.buckets = malloc(sizeof table->buckets * table->cap * 2),
	};

	js_gc_table_node_t **tasks[new_table.cap];

	for (size_t i = 0; i < sizeof table->buckets * new_table.cap; i++) {
		new_table.buckets[i] = NULL;
		tasks[i] = &new_table.buckets[i];
	}

	for (size_t i = 0; i < table->cap; i++) {
		for (js_gc_table_node_t *node = table->buckets[i]; node != NULL; ) {
			js_gc_table_node_t *curr = node;

			node = curr->next;
			curr->next = NULL;

			size_t index = bucket_i(new_table.cap, curr->node.ptr);

			*tasks[index] = curr;
			tasks[index] = &curr->next;
		}
	}

	free(table->buckets);

	*table = new_table;
}

static bool table_add_stat(js_gc_table_t *table, js_gc_node_t node, bool enlarge) {
	if (enlarge && table->n >= table->cap) table_enlarge(table);

	js_gc_table_node_t **target = &table->buckets[bucket_i(table, &node)];
	while (*target) {
		if (
			(*target)->node.type == node.type &&
			(*target)->node.ptr == node.ptr
		) return false;
		target = (*target)->next;
	}

	*target = malloc(sizeof *target);
	(*target)->next = NULL;
	(*target)->node = node;
	table->n++;

	return true;
}

static bool table_del(js_gc_table_t *table, js_gc_node_t node) {
	js_gc_table_node_t **target = &table->buckets[bucket_i(table, &node)];

	while (*target) {
		if (
			(*target)->node.type == node.type &&
			(*target)->node.ptr == node.ptr
		) {
			js_gc_table_node_t *curr = *target;
			*target = (*target)->next;
			free(curr);

			table->n--;
			return true;
		}

		target = (*target)->next;
	}

	return false;
}

static void table_free(js_gc_table_t table) {
	for (size_t i = 0; i < table.cap; i++) {
		for (js_gc_table_node_t *node = table.buckets[i]; node != NULL; ) {
			js_gc_table_node_t *curr = node;

			node = curr->next;
			free(curr);
		}
	}

	free(table.buckets);
}

static void pin_obj(js_gc_table_t *table, js_obj_t obj) {
	bool added = table_add_stat(table, (js_gc_node_t){ .ptr = obj, .type = JS_GC_OBJECT }, true);
	if (added) obj->gc.refs++;
}
static void pin_str(js_gc_table_t *table, js_gc_str_t str) {
	bool added = table_add_stat(table, (js_gc_node_t){ .ptr = str, .type = JS_GC_STRING }, true);
	if (added) str->gc.refs++;
}

static bool unpin_obj(js_gc_table_t *table, js_obj_t obj) {
	if (obj->gc.refs == 0) return true;

	bool removed = table_del(table, (js_gc_node_t){ .ptr = obj, .type = JS_GC_OBJECT });
	if (removed) obj->gc.refs--;

	return obj->gc.refs == 0;
}
static bool unpin_str(js_gc_table_t *table, js_gc_str_t str) {
	if (str->gc.refs == 0) return true;

	bool removed = table_del(table, (js_gc_node_t){ .ptr = str, .type = JS_GC_STRING });
	if (removed) str->gc.refs--;

	return str->gc.refs == 0;
}

static js_gc_task_stack_t stack_new(const js_gc_table_t *table) {
	size_t n = table->cap > 16 ? table->cap : 16;
	size_t j = 0;
	js_gc_task_stack_t stack;
	stack.cap = n;
	stack.n = table->n;
	stack.tasks = malloc(sizeof(js_gc_task_node_t) * n);

	for (size_t i = 0; i < table->cap; i++) {
		for (js_gc_table_node_t *node = table->buckets[i]; node; node = node->next) {
			stack.tasks[j++] = (js_gc_task_node_t){
				.node = node->node,
				.prev_ref = NULL,
			};
		}
	}

	return stack;
}
static void stack_free(const js_gc_task_stack_t *stack) {
	free(stack->tasks);
}

static void table_dryrun(const js_gc_table_t *table) {
	if (table->n == 0) return;

	js_gc_task_stack_t stack = stack_new(table);

	while(stack.n > 0) {
		js_gc_task_node_t task = stack.tasks[--stack.n];
		switch (task.node.type) {
			case JS_GC_OBJECT:
				js_gc_dryrun_obj(&stack, task.prev_ref, task.node.ptr);
				break;
		}
	}

	stack_free(&stack);
}
static void table_reclaim(const js_gc_table_t *table) {
	js_gc_task_stack_t stack = stack_new(table);

	while(stack.n > 0) {
		js_gc_task_node_t task = stack.tasks[--stack.n];
		switch (task.node.type) {
			case JS_GC_OBJECT:
				js_gc_reclaim_obj(&stack, task.prev_ref, task.node.ptr);
				break;
			case JS_GC_STRING:
				js_gc_reclaim_str(&stack, task.prev_ref, task.node.ptr);
				break;
		}
	}

	stack_free(&stack);
}

void js_gc_stack_push(js_gc_task_stack_t *stack, js_gc_task_node_t task) {
	while (stack->n >= stack->cap) {
		stack->cap *= 2;
		stack->tasks = realloc(stack->tasks, sizeof *stack->tasks * stack->cap);
	}

	stack->tasks[stack->n++] = task;
}
bool js_gc_stack_pop(js_gc_task_stack_t *stack, js_gc_task_node_t *pout) {
	if (stack->n == 0) return false;

	*pout = stack->tasks[--stack->n];
	return true;
}

void js_gc_pin_val(js_gc_table_t *table, js_value_t val) {
	if (val.type == JS_T_OBJ) pin_obj(table, val.data);
	if (val.type == JS_T_STR && ((js_gc_str_t)val.data)->str.dynamic) pin_str(table, val.data);
}
void js_gc_unpin_val(js_gc_table_t *table, js_value_t val) {
	if (val.type == JS_T_OBJ) unpin_obj(table, val.data);
	if (val.type == JS_T_STR && ((js_gc_str_t)val.data)->str.dynamic) unpin_str(table, val.data);
}

void js_gc_reclaim_frame(js_ctx_t ctx) {
	table_dryrun(&ctx->gc);
	table_reclaim(&ctx->gc);
	table_free(ctx->gc);
}
